#include "../common.h"

#include <netinet/ip.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/epoll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <sched.h>
#include <poll.h>

#include "../kernel/consts.h"
#include "../kernel/structs.h"
#include "../lock.h"
#include "../rwlock.h"
#include "../structs.h"
#include "../checksum.h"
#include "../ip.h"
#include "../exanic.h"
#include "../udp_queue.h"
#include "../tcp_buffer.h"
#include "../notify.h"
#include "override.h"
#include "trace.h"
#include "common.h"

void __chk_fail(void);

/* Skip skip_len bytes in the iovec before copying data
 * Returns number of bytes copied */
static inline size_t
copy_to_iovec(const struct iovec * restrict iov, size_t iovcnt,
              size_t skip_len, char *buf, size_t buf_len)
{
    size_t total_len = skip_len + buf_len;
    size_t offs, i;
    char *p;

    offs = 0;
    p = buf;
    for (i = 0; i < iovcnt && offs < total_len; i++)
    {
        size_t len = iov[i].iov_len < total_len - offs
                   ? iov[i].iov_len : total_len - offs;
        size_t skip = offs < skip_len ? skip_len - offs : 0;
        if (skip < len)
        {
            memcpy(iov[i].iov_base + skip, p, len - skip);
            p += len - skip;
        }
        offs += len;
    }

    return offs - skip_len;
}

static inline bool
__recv_block_udp_ready(struct exa_socket * restrict sock, int *ret,
                       struct exa_endpoint *ep, char **pkt, size_t *pkt_len)
{
    assert(exa_read_locked(&sock->lock));
    if (sock->state == NULL)
    {
        /* Socket no longer exists, how can this happen? */
        errno = EBADF;
        *ret = -1;
        return true;
    }
    exa_lock(&sock->state->rx_lock);
    if (exa_udp_queue_read_begin(sock, ep, pkt, pkt_len) == 0)
    {
        *ret = 0;
        return true;
    }
    else if (sock->state->rx_shutdown)
    {
        *pkt = NULL;
        *pkt_len = 0;
        *ret = 0;
        return true;
    }
    exa_unlock(&sock->state->rx_lock);
    return false;
}

/* Block until data is available in the UDP receive buffer
 * On entry, socket lock is held
 * On success, returns 0 with socket rx_lock and socket lock held
 * Otherwise returns -1 with only socket lock held */
static int
recv_block_udp(struct exa_socket * restrict sock, int fd, int flags,
               struct exa_endpoint *ep, char **pkt, size_t *pkt_len)
{
    bool nonblock = (flags & MSG_DONTWAIT) || (sock->flags & O_NONBLOCK);
    int ret;

    assert(sock->bound);
    assert(exa_read_locked(&sock->lock));

    do_socket_wait(sock, fd, nonblock, __recv_block_udp_ready,
                   ret, ep, pkt, pkt_len);
    return ret;
}

static inline bool
__recv_block_tcp_ready(struct exa_socket * restrict sock, int *ret,
                       char ** restrict buf1, size_t * restrict len1,
                       char ** restrict buf2, size_t * restrict len2)
{
    assert(exa_read_locked(&sock->lock));
    if (sock->state == NULL)
    {
        /* Socket no longer exists, how can this happen? */
        errno = EBADF;
        *ret = -1;
        return true;
    }
    exa_lock(&sock->state->rx_lock);
    if (exa_tcp_rx_buffer_read_begin(sock, buf1, len1, buf2, len2) == -1)
    {
        exa_unlock(&sock->state->rx_lock);
        errno = EIO;
        *ret = -1;
        return true;
    }
    else if (*len1 > 0 || *len2 > 0 || exa_tcp_rx_buffer_eof(sock) ||
             sock->state->rx_shutdown)
    {
        *ret = 0;
        return true;
    }
    exa_unlock(&sock->state->rx_lock);
    return false;
}

/* Block until data is available in the TCP receive buffer
 * On entry, socket lock is held
 * On success, returns 0 with socket rx_lock and socket lock held
 * Otherwise returns -1 with only socket lock held */
static int
recv_block_tcp(struct exa_socket * restrict sock, int fd, int flags,
               char ** restrict buf1, size_t * restrict len1,
               char ** restrict buf2, size_t * restrict len2)
{
    bool nonblock = (flags & MSG_DONTWAIT) || (sock->flags & O_NONBLOCK);
    int ret;

    assert(sock->bound);
    assert(exa_read_locked(&sock->lock));

    do_socket_wait(sock, fd, nonblock, __recv_block_tcp_ready,
                   ret, buf1, len1, buf2, len2);
    return ret;
}

static ssize_t
recvfrom_udp(struct exa_socket * restrict sock, int sockfd,
             void *buf, size_t len, int flags,
             struct sockaddr *src_addr, socklen_t *addrlen)
{
    struct exa_endpoint ep;
    char *pkt;
    size_t pkt_len, data_len;

    assert(exa_read_locked(&sock->lock));

    if (!sock->bound)
    {
        errno = EINVAL;
        return -1;
    }

    /* Block until packet is available, or error
     * Returns with rx_lock held */
    if (recv_block_udp(sock, sockfd, flags, &ep, &pkt, &pkt_len) == -1)
        return -1;

    data_len = pkt_len < len ? pkt_len : len;

    /* Copy data */
    memcpy(buf, pkt, data_len);

    /* Copy address */
    if (src_addr != NULL)
    {
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = ep.port.peer;
        addr.sin_addr.s_addr = ep.addr.peer;

        memcpy(src_addr, &addr,
               *addrlen < sizeof(addr) ? *addrlen : sizeof(addr));
        *addrlen = sizeof(addr);
    }

    /* Finished reading packet */
    if (flags & MSG_PEEK)
        exa_udp_queue_read_abort(sock);
    else
    {
        exa_udp_queue_read_end(sock);
        exa_notify_udp_read_update(sock);
    }

    exa_unlock(&sock->state->rx_lock);

    return (flags & MSG_TRUNC) ? pkt_len : data_len;
}

static ssize_t
recvfrom_tcp(struct exa_socket * restrict sock, int sockfd,
             void *buf, size_t len, int flags,
             struct sockaddr *src_addr, socklen_t *addrlen)
{
    char *rx_buf1, *rx_buf2;
    size_t rx_len1, rx_len2, recv_len;

    assert(exa_read_locked(&sock->lock));

    if (!sock->connected)
    {
        errno = EINVAL;
        return -1;
    }

    /* Block until packet is available, or error */
    if (recv_block_tcp(sock, sockfd, flags, &rx_buf1, &rx_len1, &rx_buf2,
                       &rx_len2) == -1)
        return -1;

    if (rx_len2 == 0 || len <= rx_len1)
    {
        /* Read does not cross receive buffer wrap */
        recv_len = rx_len1 < len ? rx_len1 : len;
        memcpy(buf, rx_buf1, recv_len);
    }
    else
    {
        /* Read crosses receive buffer wrap */
        recv_len = (rx_len1 + rx_len2) < len ? (rx_len1 + rx_len2) : len;
        memcpy(buf, rx_buf1, rx_len1);
        memcpy(buf + rx_len1, rx_buf2, recv_len - rx_len1);
    }
    exa_tcp_rx_buffer_read_end(sock, (flags & MSG_PEEK) ? 0 : recv_len);
    exa_notify_tcp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    /* Copy address */
    if (src_addr != NULL)
    {
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = sock->bind.ip.port.peer;
        addr.sin_addr.s_addr = sock->bind.ip.addr.peer;

        memcpy(src_addr, &addr,
               *addrlen < sizeof(addr) ? *addrlen : sizeof(addr));
        *addrlen = sizeof(addr);
    }

    return recv_len;
}

static ssize_t
recvfrom_bypass(struct exa_socket * restrict sock, int sockfd,
                void *buf, size_t len, int flags,
                struct sockaddr *src_addr, socklen_t *addrlen)
{
    assert(exa_read_locked(&sock->lock));

    if (sock->domain == AF_INET && sock->type == SOCK_DGRAM)
        return recvfrom_udp(sock, sockfd, buf, len, flags, src_addr, addrlen);
    else if (sock->domain == AF_INET && sock->type == SOCK_STREAM)
        return recvfrom_tcp(sock, sockfd, buf, len, flags, src_addr, addrlen);
    else
    {
        errno = EINVAL;
        return -1;
    }
}

__attribute__((visibility("default")))
ssize_t
recv(int sockfd, void *buf, size_t len, int flags)
{
    struct exa_socket * restrict sock = exa_socket_get(sockfd);
    ssize_t ret;

    TRACE_CALL("recv");
    TRACE_ARG(INT, sockfd);
    TRACE_FLUSH();

    if (sock == NULL)
        ret = libc_recv(sockfd, buf, len, flags);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_recv(sockfd, buf, len, flags);
        }
        else
        {
            ret = recvfrom_bypass(sock, sockfd, buf, len, flags, NULL, NULL);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_ARG(LONG, len);
    TRACE_LAST_ARG(BITS, flags, msg_flags);
    TRACE_RETURN(LONG, ret);

    return ret;
}

__attribute__((visibility("default")))
ssize_t
__recv_chk(int sockfd, void *buf, size_t len, size_t buflen, int flags)
{
    struct exa_socket * restrict sock = exa_socket_get(sockfd);
    ssize_t ret;

    TRACE_CALL("__recv_chk");
    TRACE_ARG(INT, sockfd);
    TRACE_FLUSH();

    if (buflen < len)
        __chk_fail();

    if (sock == NULL)
        ret = libc_recv(sockfd, buf, len, flags);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_recv(sockfd, buf, len, flags);
        }
        else
        {
            ret = recvfrom_bypass(sock, sockfd, buf, len, flags, NULL, NULL);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_ARG(LONG, len);
    TRACE_ARG(LONG, buflen);
    TRACE_LAST_ARG(BITS, flags, msg_flags);
    TRACE_RETURN(LONG, ret);

    return ret;
}

__attribute__((visibility("default")))
ssize_t
recvfrom(int sockfd, void *buf, size_t len, int flags,
         struct sockaddr *src_addr, socklen_t *addrlen)
{
    struct exa_socket * restrict sock = exa_socket_get(sockfd);
    ssize_t ret;

    TRACE_CALL("recvfrom");
    TRACE_ARG(INT, sockfd);
    TRACE_FLUSH();

    if (sock == NULL)
        ret = libc_recvfrom(sockfd, buf, len, flags, src_addr, addrlen);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_recvfrom(sockfd, buf, len, flags, src_addr, addrlen);
        }
        else
        {
            ret = recvfrom_bypass(sock, sockfd, buf, len, flags, src_addr,
                                  addrlen);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_ARG(LONG, len);
    TRACE_ARG(BITS, flags, msg_flags);
    TRACE_ARG(SOCKADDR_PTR, src_addr);
    TRACE_LAST_ARG(INT_PTR, addrlen);
    TRACE_RETURN(LONG, ret);

    return ret;
}

__attribute__((visibility("default")))
ssize_t
__recvfrom_chk(int sockfd, void *buf, size_t len, size_t buflen,
               int flags, struct sockaddr *src_addr, socklen_t *addrlen)
{
    struct exa_socket * restrict sock = exa_socket_get(sockfd);
    ssize_t ret;

    TRACE_CALL("__recvfrom_chk");
    TRACE_ARG(INT, sockfd);
    TRACE_FLUSH();

    if (buflen < len)
        __chk_fail();

    if (sock == NULL)
        ret = libc_recvfrom(sockfd, buf, len, flags, src_addr, addrlen);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_recvfrom(sockfd, buf, len, flags, src_addr, addrlen);
        }
        else
        {
            ret = recvfrom_bypass(sock, sockfd, buf, len, flags, src_addr,
                                  addrlen);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_ARG(LONG, len);
    TRACE_ARG(LONG, buflen);
    TRACE_ARG(BITS, flags, msg_flags);
    TRACE_ARG(SOCKADDR_PTR, src_addr);
    TRACE_LAST_ARG(INT_PTR, addrlen);
    TRACE_RETURN(LONG, ret);

    return ret;
}

static ssize_t
recvmsg_udp(struct exa_socket * restrict sock, int sockfd,
            struct msghdr *msg, int flags)
{
    struct exa_endpoint ep;
    char *pkt;
    size_t pkt_len, recv_len;

    assert(exa_read_locked(&sock->lock));

    /* Block until packet is available, or error */
    if (recv_block_udp(sock, sockfd, flags, &ep, &pkt, &pkt_len) == -1)
        return -1;

    /* Copy data */
    recv_len = copy_to_iovec(msg->msg_iov, msg->msg_iovlen, 0, pkt, pkt_len);

    /* Copy address */
    if (msg->msg_name != NULL)
    {
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = ep.port.peer;
        addr.sin_addr.s_addr = ep.addr.peer;

        memcpy(msg->msg_name, &addr, msg->msg_namelen < sizeof(addr)
               ? msg->msg_namelen : sizeof(addr));
    }

    /* TODO: Populate ancillary data */
    if (msg->msg_control != NULL)
        msg->msg_controllen = 0;

    /* Populate flags */
    msg->msg_flags = (recv_len < pkt_len ? MSG_TRUNC : 0);

    /* Finished reading packet */
    if (flags & MSG_PEEK)
        exa_udp_queue_read_abort(sock);
    else
    {
        exa_udp_queue_read_end(sock);
        exa_notify_udp_read_update(sock);
    }

    exa_unlock(&sock->state->rx_lock);

    return (flags & MSG_TRUNC) ? pkt_len : recv_len;
}

static ssize_t
recvmsg_tcp(struct exa_socket * restrict sock, int sockfd,
            struct msghdr *msg, int flags)
{
    char *rx_buf1, *rx_buf2;
    size_t rx_len1, rx_len2, recv_len;

    assert(exa_read_locked(&sock->lock));

    if (!sock->connected)
    {
        errno = EINVAL;
        return -1;
    }

    /* Block until packet is available, or error */
    if (recv_block_tcp(sock, sockfd, flags, &rx_buf1, &rx_len1, &rx_buf2,
                       &rx_len2) == -1)
        return -1;

    /* Copy data */
    recv_len = copy_to_iovec(msg->msg_iov, msg->msg_iovlen, 0, rx_buf1,
                             rx_len1);
    if (rx_len2 != 0 && recv_len == rx_len1)
        recv_len += copy_to_iovec(msg->msg_iov, msg->msg_iovlen, rx_len1,
                                rx_buf2, rx_len2);

    /* Copy address */
    if (msg->msg_name != NULL)
    {
        struct sockaddr_in addr;

        addr.sin_family = AF_INET;
        addr.sin_port = sock->bind.ip.port.peer;
        addr.sin_addr.s_addr = sock->bind.ip.addr.peer;

        memcpy(msg->msg_name, &addr, msg->msg_namelen < sizeof(addr)
               ? msg->msg_namelen : sizeof(addr));
    }

    /* TODO: Populate ancillary data */
    if (msg->msg_control != NULL)
        msg->msg_controllen = 0;

    /* Populate flags */
    msg->msg_flags = 0;

    exa_tcp_rx_buffer_read_end(sock, (flags & MSG_PEEK) ? 0 : recv_len);
    exa_notify_tcp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    return recv_len;
}

static ssize_t
recvmsg_bypass(struct exa_socket * restrict sock, int sockfd,
               struct msghdr *msg, int flags)
{
    assert(exa_read_locked(&sock->lock));

    if (sock->domain == AF_INET && sock->type == SOCK_DGRAM)
        return recvmsg_udp(sock, sockfd, msg, flags);
    else if (sock->domain == AF_INET && sock->type == SOCK_STREAM)
        return recvmsg_tcp(sock, sockfd, msg, flags);
    else
    {
        errno = EINVAL;
        return -1;
    }
}

__attribute__((visibility("default")))
ssize_t
recvmsg(int sockfd, struct msghdr *msg, int flags)
{
    struct exa_socket * restrict sock = exa_socket_get(sockfd);
    ssize_t ret;

    TRACE_CALL("recvmsg");
    TRACE_ARG(INT, sockfd);
    TRACE_FLUSH();

    if (sock == NULL)
        ret = libc_recvmsg(sockfd, msg, flags);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_recvmsg(sockfd, msg, flags);
        }
        else
        {
            ret = recvmsg_bypass(sock, sockfd, msg, flags);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(MSG_PTR, msg, ret);
    TRACE_LAST_ARG(BITS, flags, msg_flags);
    TRACE_RETURN(LONG, ret);

    return ret;
}

static ssize_t
read_udp(struct exa_socket * restrict sock, int fd, void *buf, size_t count)
{
    char *pkt;
    size_t pkt_len, data_len;

    assert(exa_read_locked(&sock->lock));

    /* Block until packet is available, or error */
    if (recv_block_udp(sock, fd, 0, NULL, &pkt, &pkt_len) == -1)
        return -1;

    data_len = pkt_len < count ? pkt_len : count;

    /* Copy data */
    memcpy(buf, pkt, data_len);

    /* Remainder of packet is discarded */
    exa_udp_queue_read_end(sock);
    exa_notify_udp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    return data_len;
}

static ssize_t
read_tcp(struct exa_socket * restrict sock, int fd, void *buf, size_t count)
{
    char *rx_buf1, *rx_buf2;
    size_t rx_len1, rx_len2, data_len;

    assert(exa_read_locked(&sock->lock));

    if (!sock->connected)
    {
        errno = ENOTCONN;
        return -1;
    }

    if (count == 0)
        return 0;

    /* Block until data is available, or error */
    if (recv_block_tcp(sock, fd, 0, &rx_buf1, &rx_len1, &rx_buf2,
                       &rx_len2) == -1)
        return -1;

    if (rx_len2 == 0 || count <= rx_len1)
    {
        /* Read does not cross receive buffer wrap */
        data_len = rx_len1 < count ? rx_len1 : count;
        memcpy(buf, rx_buf1, data_len);
    }
    else
    {
        /* Read crosses receive buffer wrap */
        data_len = (rx_len1 + rx_len2) < count ? (rx_len1 + rx_len2) : count;
        memcpy(buf, rx_buf1, rx_len1);
        memcpy(buf + rx_len1, rx_buf2, data_len - rx_len1);
    }
    exa_tcp_rx_buffer_read_end(sock, data_len);
    exa_notify_tcp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    return data_len;
}

static ssize_t
read_bypass(struct exa_socket * restrict sock, int fd, void *buf, size_t count)
{
    assert(exa_read_locked(&sock->lock));

    if (sock->domain == AF_INET && sock->type == SOCK_DGRAM)
        return read_udp(sock, fd, buf, count);
    else if (sock->domain == AF_INET && sock->type == SOCK_STREAM)
        return read_tcp(sock, fd, buf, count);
    else
    {
        errno = EINVAL;
        return -1;
    }
}

__attribute__((visibility("default")))
ssize_t
read(int fd, void *buf, size_t count)
{
    struct exa_socket * restrict sock = exa_socket_get(fd);
    ssize_t ret;

    TRACE_CALL("read");
    TRACE_ARG(INT, fd);
    TRACE_FLUSH();

    if (sock == NULL)
        ret = libc_read(fd, buf, count);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_read(fd, buf, count);
        }
        else
        {
            ret = read_bypass(sock, fd, buf, count);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_LAST_ARG(LONG, count);
    TRACE_RETURN(LONG, ret);

    return ret;
}

__attribute__((visibility("default")))
ssize_t
__read_chk(int fd, void *buf, size_t nbytes, size_t buflen)
{
    struct exa_socket * restrict sock = exa_socket_get(fd);
    ssize_t ret;

    TRACE_CALL("__read_chk");
    TRACE_ARG(INT, fd);
    TRACE_FLUSH();

    if (buflen < nbytes)
        __chk_fail();

    if (sock == NULL)
        ret = libc_read(fd, buf, nbytes);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_read(fd, buf, nbytes);
        }
        else
        {
            ret = read_bypass(sock, fd, buf, nbytes);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(BUF, buf, ret);
    TRACE_ARG(LONG, nbytes);
    TRACE_LAST_ARG(LONG, buflen);
    TRACE_RETURN(LONG, ret);

    return ret;
}

static ssize_t
readv_udp(struct exa_socket * restrict sock, int fd, const struct iovec *iov,
          size_t iovcnt)
{
    char *pkt;
    size_t pkt_len, data_len;

    assert(exa_read_locked(&sock->lock));

    /* Block until packet is available, or error */
    if (recv_block_udp(sock, fd, 0, NULL, &pkt, &pkt_len) == -1)
        return -1;

    /* Copy data */
    data_len = copy_to_iovec(iov, iovcnt, 0, pkt, pkt_len);

    /* Remainder of packet is discarded */
    exa_udp_queue_read_end(sock);
    exa_notify_udp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    return data_len;
}

static ssize_t
readv_tcp(struct exa_socket * restrict sock, int fd, const struct iovec *iov,
          size_t iovcnt)
{
    char *rx_buf1, *rx_buf2;
    size_t rx_len1, rx_len2, data_len;

    assert(exa_read_locked(&sock->lock));

    if (!sock->connected)
    {
        errno = ENOTCONN;
        return -1;
    }

    /* Block until data is available, or error */
    if (recv_block_tcp(sock, fd, 0, &rx_buf1, &rx_len1, &rx_buf2,
                       &rx_len2) == -1)
        return -1;

    /* Copy data */
    data_len = copy_to_iovec(iov, iovcnt, 0, rx_buf1, rx_len1);
    if (rx_len2 != 0 && data_len == rx_len1)
        data_len += copy_to_iovec(iov, iovcnt, rx_len1, rx_buf2, rx_len2);

    exa_tcp_rx_buffer_read_end(sock, data_len);
    exa_notify_tcp_read_update(sock);
    exa_unlock(&sock->state->rx_lock);

    return data_len;
}

static ssize_t
readv_bypass(struct exa_socket * restrict sock, int fd, const struct iovec *iov,
             size_t iovcnt)
{
    assert(exa_read_locked(&sock->lock));

    if (sock->domain == AF_INET && sock->type == SOCK_DGRAM)
        return readv_udp(sock, fd, iov, iovcnt);
    else if (sock->domain == AF_INET && sock->type == SOCK_STREAM)
        return readv_tcp(sock, fd, iov, iovcnt);
    else
    {
        errno = EINVAL;
        return -1;
    }
}

__attribute__((visibility("default")))
ssize_t
readv(int fd, const struct iovec *iov, int iovcnt)
{
    struct exa_socket * restrict sock = exa_socket_get(fd);
    ssize_t ret;

    TRACE_CALL("readv");
    TRACE_ARG(INT, fd);
    TRACE_FLUSH();

    if (sock == NULL)
        ret = libc_readv(fd, iov, iovcnt);
    else
    {
        exa_read_lock(&sock->lock);

        if (!sock->bypass)
        {
            exa_read_unlock(&sock->lock);
            ret = libc_readv(fd, iov, iovcnt);
        }
        else
        {
            ret = readv_bypass(sock, fd, iov, iovcnt);
            exa_read_unlock(&sock->lock);
        }
    }

    TRACE_ARG(IOVEC_ARRAY, iov, iovcnt, ret);
    TRACE_LAST_ARG(INT, iovcnt);
    TRACE_RETURN(LONG, ret);

    return ret;
}
