#include <math.h>
#include <stdio.h>
#include <syslog.h>

#include <exanic/exanic.h>
#include <exanic/pcie_if.h>
#include <exanic/register.h>
#include <exanic/time.h>

#include "common.h"
#include "exanic.h"

#define LEN 4


struct exanic_exanic_sync_state
{
    exanic_t *exanic;
    exanic_t *exanic_src;
    uint64_t time_ns;   /* Time of last measurement (ns since epoch) */
    int64_t offset_ns;  /* Last measured offset of clock (ns) */
    double adj;         /* Currently applied adjustment value */
    double drift[LEN];  /* Estimated drift of the underlying clock */
    int n;              /* Index into next entry in drift array */
    int startup_mode;   /* Nonzero if measurement history is incomplete */
    int log_next;       /* Make sure next measurement is logged */
    time_t last_log;    /* Time of last log message (s since epoch) */
};


/* Average measurements to estimate underlying clock drift */
static double calc_drift(struct exanic_exanic_sync_state *state)
{
    double drift;
    int i;

    if (state->startup_mode)
    {
        drift = 0;
        for (i = 0; i < state->n; i++)
            drift += state->drift[i];
        if (state->n > 0)
            drift /= state->n;
    }
    else
    {
        drift = 0;
        for (i = 0; i < LEN; i++)
            drift += state->drift[i];
        drift /= LEN;
    }

    return drift;
}


/* Get current time from both exanics */
static void get_current_time(struct exanic_exanic_sync_state *state,
        uint64_t *exanic_time_ns, uint64_t *src_time_ns)
{
    uint32_t exanic_tick, src_tick, exanic_tick_2, src_tick_2;

    /* Read twice in different order to cancel out latency */
    asm volatile ("mfence");
    exanic_tick = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
    asm volatile ("mfence");
    src_tick = exanic_register_read(state->exanic_src,
            REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
    asm volatile ("mfence");
    src_tick_2 = exanic_register_read(state->exanic_src,
            REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
    asm volatile ("mfence");
    exanic_tick_2 = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
    asm volatile ("mfence");

    exanic_tick += (exanic_tick_2 - exanic_tick) / 2;
    src_tick += (src_tick_2 - src_tick) / 2;

    *exanic_time_ns = exanic_timestamp_to_counter(state->exanic, exanic_tick);
    *src_time_ns = exanic_timestamp_to_counter(state->exanic_src, src_tick);
}


struct exanic_exanic_sync_state *init_exanic_exanic_sync(exanic_t *exanic_sync,
        exanic_t *exanic_src)
{
    struct exanic_exanic_sync_state *state;
    uint64_t src_time_ns, exanic_time_ns;
    uint32_t exanic_adj;

    state = malloc(sizeof(struct exanic_exanic_sync_state));

    exanic_retain_handle(exanic_sync);
    exanic_retain_handle(exanic_src);

    state->exanic = exanic_sync;
    state->exanic_src = exanic_src;

    log_printf(LOG_INFO, "%s: Starting clock discipline using %s clock",
            state->exanic->name, state->exanic_src->name);

    /* Set up state struct with current time */
    get_current_time(state, &exanic_time_ns, &src_time_ns);
    state->time_ns = src_time_ns;
    state->offset_ns = (int64_t)(exanic_time_ns - src_time_ns);
    state->startup_mode = 1;
    state->log_next = 0;
    state->last_log = src_time_ns / 1000000000;

    /* Get current adjustment from exanic */
    exanic_adj = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ));
    state->adj = from_exanic_clk_adj(exanic_adj);

    /* Use current adjustment as our initial estimate of drift */
    state->drift[0] = - state->adj;
    state->n = 1;

    return state;
}


enum sync_status poll_exanic_exanic_sync(struct exanic_exanic_sync_state *state)
{
    uint64_t src_time_ns, exanic_time_ns;
    int64_t offset_ns, expected_offset_ns;
    time_t current_time;
    double drift, adj;
    uint32_t exanic_adj;
    int i;
    int fast_poll = 0;

    /* Calculate underlying clock drift */
    drift = calc_drift(state);

    /* Do measurements.  This is repeated up to 3 times if we get a
     * measurement that is too far from the expected value. */
    for (i = 0; i < 3; i++)
    {
        /* Get current time from both exanics and calculate actual offset */
        get_current_time(state, &exanic_time_ns, &src_time_ns);
        offset_ns = (int64_t)(exanic_time_ns - src_time_ns);
        current_time = src_time_ns / 1000000000;

        /* Calculate expected offset based on the current adjustment and
         * estimated drift */
        expected_offset_ns = state->offset_ns +
            (drift + state->adj) * (src_time_ns - state->time_ns);

        /* Accept measurement if it is within 10us of expected value */
        if (expected_offset_ns - 10000 < offset_ns &&
                offset_ns < expected_offset_ns + 10000)
            break;
    }

    if (i == 3)
    {
        /* Measurement is far from the expected value, so start polling
         * at a faster rate until things stabilise */
        fast_poll = 1;
        state->log_next = 1;
    }

    /* Reset clock if offset is more than 1ms */
    if (offset_ns > 1000000 || offset_ns < -1000000)
    {
        uint32_t time_tick;

        if (!state->startup_mode || state->n > 1)
            log_printf(LOG_WARNING, "%s: Clock error exceeds limits, "
                    "resetting clock: %ld ns",
                    state->exanic->name, offset_ns);

        time_tick = exanic_counter_to_timestamp(state->exanic, src_time_ns);
        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ), 0);
        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_SET), time_tick);
        state->time_ns = src_time_ns;
        state->offset_ns = 0;
        state->adj = 0;
        state->n = 0;
        state->startup_mode = 1;
        state->log_next = 0;
        state->last_log = current_time;
        return SYNC_FAST_POLL;
    }

    /* Record drift measurement */
    state->drift[state->n] = drift + (double)(offset_ns - expected_offset_ns) /
        (src_time_ns - state->time_ns);
    if (++state->n >= LEN)
        state->n = state->startup_mode = 0;

    /* Set adjustment to compensate for drift and to correct offset */
    adj = - drift - offset_ns / (1000000000 * POLL_INTERVAL);
    exanic_adj = to_exanic_clk_adj(adj);

    exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
            exanic_adj);

    /* Store measurements and current adjustment */
    state->time_ns = src_time_ns;
    state->offset_ns = offset_ns;
    state->adj = from_exanic_clk_adj(exanic_adj);

    /* Print status */
    if (state->log_next || state->last_log + LOG_INTERVAL <= current_time)
    {
        log_printf(LOG_INFO, "%s: Clock offset from %s: %.3f us "
                " drift: %.3f ppm", state->exanic->name,
                state->exanic_src->name,
                offset_ns * 0.001, drift * 1000000);
        state->last_log = current_time;

        /* Log again if offset is more than 2us */
        state->log_next = (offset_ns < -2000 || offset_ns > 2000);
    }

    return fast_poll ? SYNC_FAST_POLL : SYNC_OK;
}


void shutdown_exanic_exanic_sync(struct exanic_exanic_sync_state *state)
{
    uint32_t exanic_adj;
    double drift;

    log_printf(LOG_INFO, "%s: Stopping clock discipline using %s clock",
            state->exanic->name, state->exanic_src->name);

    /* Set adjustment to compensate for drift only */
    drift = calc_drift(state);
    exanic_adj = to_exanic_clk_adj(-drift);
    exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
            exanic_adj);

    exanic_release_handle(state->exanic);
    exanic_release_handle(state->exanic_src);

    free(state);
}
