#include <math.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/time.h>

#include <exanic/exanic.h>
#include <exanic/pcie_if.h>
#include <exanic/register.h>
#include <exanic/util.h>

#include "common.h"
#include "pps.h"

#define LEN 8
#define SKIP_MAX 5
#define SPURIOUS_MAX 3
#define TIMEOUT_SECONDS 10


struct exanic_pps_sync_state
{
    exanic_t *exanic;
    uint32_t tick_hz;   /* Nominal frequency of clock (ticks/second) */
    uint64_t poll_time_us; /* Time of last poll (us since epoch) */
    uint32_t pps_reg;   /* Last seen value of the PPS register */
    time_t pps_time;    /* Time of last PPS pulse (seconds since epoch) */
    int32_t pps_offset; /* Offset at last PPS pulse (ticks) */
    double adj;         /* Currently applied adjustment value */
    double drift[LEN];  /* Estimated drift of the underlying clock */
    int n;              /* Index into next entry in drift array */
    int startup_mode;   /* Nonzero if measurement history is incomplete */
    int pps_signal;     /* 0 = no signal, 1 = signal, -1 = indeterminate */
    int log_next;       /* Make sure next measurement is logged */
    uint64_t last_log_us; /* Time of last log message (us since epoch) */
    time_t pps_last_seen; /* Time of last poll that a PPS pulse was detected */
    int spurious_pps_count; /* Number of consecutive spurious PPS pulses */
};


/* Average measurements to estimate underlying clock drift */
static double calc_drift(struct exanic_pps_sync_state *state)
{
    double drift;
    int i;

    if (state->startup_mode)
    {
        drift = 0;
        for (i = 0; i < state->n; i++)
            drift += state->drift[i];
        if (state->n > 0)
            drift /= state->n;
    }
    else
    {
        drift = 0;
        for (i = 0; i < LEN; i++)
            drift += state->drift[i];
        drift /= LEN;
    }

    return drift;
}


struct exanic_pps_sync_state *init_exanic_pps_sync(exanic_t *exanic,
        enum pps_type pps_type)
{
    struct exanic_pps_sync_state *state;
    struct timeval tv;
    uint32_t exanic_adj;
    exanic_hardware_id_t hw_id;

    state = malloc(sizeof(struct exanic_pps_sync_state));

    exanic_retain_handle(exanic);

    state->exanic = exanic;
    state->tick_hz = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_CLK_HZ));

    /* Check for invalid settings */
    hw_id = exanic_get_hw_type(exanic);
    if ((hw_id == EXANIC_HW_Z1 || hw_id == EXANIC_HW_Z10) &&
            pps_type == PPS_SINGLE_ENDED)
    {
        log_printf(LOG_WARNING, "%s: %s does not support single-ended PPS input",
                state->exanic->name, exanic_hardware_id_str(hw_id));
        pps_type = PPS_DIFFERENTIAL;
    }

    log_printf(LOG_INFO, "%s: Starting clock discipline using PPS",
            state->exanic->name);
    log_printf(LOG_INFO, "%s: Nominal frequency: %u Hz",
            state->exanic->name, state->tick_hz);
    log_printf(LOG_INFO, "%s: Using %s PPS input", state->exanic->name,
            pps_type == PPS_SINGLE_ENDED ? "single-ended" : "differential");

    /* PPS settings */
    if (hw_id == EXANIC_HW_X4)
    {
        uint32_t reg;

        reg = exanic_register_read(state->exanic, REG_HW_INDEX(REG_HW_SERIAL_PPS));
        if (pps_type == PPS_SINGLE_ENDED)
            reg |= EXANIC_HW_SERIAL_PPS_SINGLE;
        else
            reg &= ~EXANIC_HW_SERIAL_PPS_SINGLE;
        exanic_register_write(state->exanic, REG_HW_INDEX(REG_HW_SERIAL_PPS), reg);
    }

    /* Set up state struct */
    gettimeofday(&tv, NULL);
    state->poll_time_us = tv.tv_sec * (uint64_t)1000000 + tv.tv_usec;

    state->pps_reg = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_PPS_TIMESTAMP));
    state->pps_time = 0;
    state->pps_offset = 0;

    /* Get current adjustment from exanic */
    exanic_adj = exanic_register_read(state->exanic,
            REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ));
    state->adj = from_exanic_clk_adj(exanic_adj);

    state->n = 0;
    state->startup_mode = 1;
    state->pps_signal = -1;
    state->log_next = 1;
    state->last_log_us = state->poll_time_us;
    state->pps_last_seen = tv.tv_sec;
    state->spurious_pps_count = 0;

    return state;
}


enum sync_status poll_exanic_pps_sync(struct exanic_pps_sync_state *state)
{
    struct timeval tv;
    uint32_t pps_reg;
    uint64_t poll_time_us;
    double drift;

    gettimeofday(&tv, NULL);
    poll_time_us = tv.tv_sec * (uint64_t)1000000 + tv.tv_usec;

    /* Get clock drift from measurement history */
    drift = calc_drift(state);

    /* Calculate the time of the last PPS pulse */
    pps_reg = exanic_register_read(state->exanic, REG_EXANIC_INDEX(
                REG_EXANIC_PPS_TIMESTAMP));
    if (pps_reg != state->pps_reg)
    {
        uint64_t time_midpoint_us, time_tick;
        int32_t pps_offset;
        time_t time_sec;
        int valid_pps;

        /* Find the closest second boundary to the midpoint between this poll
         * and last poll */
        time_midpoint_us = (poll_time_us + state->poll_time_us) / 2;
        if (time_midpoint_us % 1000000 < 500000)
            time_sec = time_midpoint_us / 1000000;
        else
            time_sec = time_midpoint_us / 1000000 + 1;

        /* Get offset at PPS pulse */
        time_tick = time_sec * state->tick_hz;
        pps_offset = (int32_t)(pps_reg - time_tick);

        /* Heuristics to reject spurious PPS pulses */
        valid_pps = 0;
        /* If offset small (less than 1ms), assume non-spurious */
        if (abs(pps_offset) < state->tick_hz / 1000)
            valid_pps = 1;
        /* Assume non-spurious for first PPS pulse on startup */
        if (state->pps_time == 0)
            valid_pps = 1;
        /* If too many "spurious" pulses, then we are probably wrong */
        if (state->spurious_pps_count >= SPURIOUS_MAX)
            valid_pps = 1;

        if (!valid_pps)
        {
            state->pps_reg = pps_reg;
            state->spurious_pps_count++;
            log_printf(LOG_WARNING, "%s: Ignoring possible spurious PPS, "
                    "clock offset: %.3f us", state->exanic->name,
                    pps_offset * 1000000.0 / state->tick_hz);
        }
        else
        {
            state->spurious_pps_count = 0;

            if ((time_sec - state->pps_time) <= (SKIP_MAX + 1))
            {
                /* Measure drift */
                unsigned seconds = time_sec - state->pps_time;
                int32_t exanic_ticks = pps_reg - state->pps_reg;
                int32_t tick_diff = exanic_ticks - (seconds * state->tick_hz);
                double expected_tick_diff =
                    (drift + state->adj) * seconds * state->tick_hz;

                /* Record drift measurement */
                state->drift[state->n] = drift + (tick_diff - expected_tick_diff)
                    / (seconds * state->tick_hz);
                if (++state->n >= LEN)
                    state->n = state->startup_mode = 0;

                /* Update estimate of clock drift */
                drift = calc_drift(state);
            }

            state->pps_offset = pps_offset;
            state->pps_reg = pps_reg;
            state->pps_time = time_sec;

            if (state->pps_signal != 1)
                log_printf(LOG_INFO, "%s: PPS signal detected",
                        state->exanic->name);
            state->pps_signal = 1;
            state->pps_last_seen = tv.tv_sec;

            /* If offset is more than 1us, make sure status is logged */
            if (abs(state->pps_offset) > state->tick_hz / 1000000)
                state->log_next = 1;

            /* Log measurement */
            if (state->log_next ||
                    state->last_log_us + LOG_INTERVAL * 1000000 < poll_time_us)
            {
                log_printf(LOG_INFO, "%s: Clock offset at PPS pulse: %.3f us "
                        " drift: %.3f ppm", state->exanic->name,
                        state->pps_offset * 1000000.0 / state->tick_hz,
                        drift * 1000000);
                state->last_log_us = poll_time_us;
            }

            /* Slow down logging if offset is less than 1us */
            if (abs(state->pps_offset) < state->tick_hz / 1000000)
                state->log_next = 0;
        }
    }

    /* Check if we haven't seen a PPS pulse for a while */
    if (tv.tv_sec - state->pps_last_seen > TIMEOUT_SECONDS)
    {
        if (state->pps_signal != 0 ||
                state->last_log_us + LOG_INTERVAL * 1000000 < poll_time_us)
        {
            log_printf(LOG_WARNING, "%s: PPS signal lost",
                    state->exanic->name);
            state->last_log_us = poll_time_us;
        }

        state->pps_signal = 0;
        state->log_next = 1;
    }

    /* Reset exanic clock if offset is too large (more than 1ms) */
    if (fabs(state->pps_offset) > 0.001 * state->tick_hz)
    {
        uint32_t clk_reg;

        if (!state->startup_mode || state->n > 1)
            log_printf(LOG_WARNING, "%s: Clock error exceeds limits, "
                    "resetting clock: %d ticks",
                    state->exanic->name, state->pps_offset);

        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ), 0);
        clk_reg = exanic_register_read(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_HW_TIME));
        exanic_register_write(state->exanic,
                REG_EXANIC_INDEX(REG_EXANIC_CLK_SET),
                clk_reg - state->pps_offset);
        state->pps_offset = 0;
        state->pps_time = 0;
        state->adj = 0;
        state->n = 0;
        state->startup_mode = 1;
        state->log_next = 1;
        state->last_log_us = poll_time_us;
        state->spurious_pps_count = 0;
        return SYNC_OK;
    }

    if (!state->startup_mode || state->n > 0)
    {
        uint32_t exanic_adj;
        double offset_correction;
        int32_t time_since_pps_us;

        /* Measure time since last PPS pulse */
        time_since_pps_us = poll_time_us - state->pps_time * 1000000;
        if (time_since_pps_us < 10 * PPS_POLL_INTERVAL * 1000000)
            /* Add offset correction */
            offset_correction = - state->pps_offset /
                (10 * PPS_POLL_INTERVAL * state->tick_hz);
        else
            /* Don't add offset correction, it should be corrected by now */
            offset_correction = 0;

        /* Set adjustment to compensate for drift and to correct offset */
        state->adj = offset_correction - drift;
        exanic_adj = to_exanic_clk_adj(state->adj);
        exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
                exanic_adj);
    }

    state->poll_time_us = poll_time_us;

    return SYNC_OK;
}


void shutdown_exanic_pps_sync(struct exanic_pps_sync_state *state)
{
    uint32_t exanic_adj;
    double drift;

    log_printf(LOG_INFO, "%s: Stopping clock discipline using PPS",
            state->exanic->name);

    /* Set adjustment to compensate for drift only */
    drift = calc_drift(state);
    exanic_adj = to_exanic_clk_adj(-drift);
    exanic_register_write(state->exanic, REG_EXANIC_INDEX(REG_EXANIC_CLK_ADJ),
            exanic_adj);

    exanic_release_handle(state->exanic);

    free(state);
}
